import Image, struct

def isBitSet(n, numBit):
    return bool(n & (2 ** numBit))

def bits(n, fromBit, toBit):
    result = 0
    for i in xrange(fromBit, toBit):
        result += int((2**(i-fromBit))*isBitSet(n, i))
    return result

def rgb565toColor(n):
    return (bits(n, 0, 5)*8, bits(n, 5, 11)*4, bits(n, 11, 16)*8)

class RasterImage:
    def read(self, stream):
        fmt = '6L'
        self.type,\
        self.width,\
        self.height,\
        self.mipmapct,\
        self.pixelwidth,\
        self.pixelformat = struct.unpack(fmt, stream.read(struct.calcsize(fmt)))
        self.maps = []
        for i in xrange(self.mipmapct):
            fmt = 'L'
            blocksize = struct.unpack(fmt, stream.read(struct.calcsize(fmt)))[0]
            self.maps += [Image.new('RGBA', (self.width, self.height))]
            if self.pixelformat == 0x15:
                fmt = '4B'
                for j in xrange(blocksize/4):
                    self.maps[i].putpixel((j % self.width, j/self.width), struct.unpack(fmt, stream.read(4)))
            elif self.pixelformat in (0x32545844, 0x33545844): # DXT2/3
                fmt = 'QHHL'
                x, y = 0, 0
                for j in xrange(blocksize/16):
                    alpha,\
                    color0,\
                    color1,\
                    pixels = struct.unpack(fmt, stream.read(16))
                    colors = [rgb565toColor(color0),
                              rgb565toColor(color1)]
                    colors += [((2 * colors[0][0] + colors[1][0]) / 3, (2 * colors[0][1] + colors[1][1]) / 3, (2 * colors[0][2] + colors[1][2]) / 3)]
                    colors += [((2 * colors[1][0] + colors[0][0]) / 3, (2 * colors[1][1] + colors[0][1]) / 3, (2 * colors[1][2] + colors[0][2]) / 3)]
                    for k in xrange(16):
                        pixel = colors[bits(pixels, k*2, (k*2)+2)]
                        pixel += (bits(alpha, k*4, (k*4)+4)*16,)
                        self.maps[i].putpixel((x+(k%4), y+(k/4)), pixel)
                    x += 4
                    if x >= self.width:
                        y += 4
                        x = 0
            elif self.pixelformat in (0x34545844, 0x35545844): # DXT4/5
                fmt = 'BB3HHHL'
                x, y = 0, 0
                for j in xrange(blocksize/16):
                    alpha0,\
                    alpha1,\
                    alphaA,\
                    alphaB,\
                    alphaC,\
                    color0,\
                    color1,\
                    pixels = struct.unpack(fmt, stream.read(16))
                    alpha = alphaA+(alphaB<<16)+(alphaC<<32)
                    alphas = [alpha0, alpha1]+[None]*6
                    if alpha0 > alpha1:
                        alphas[2] = (6 * alpha0 + 1 * alpha1) / 7
                        alphas[3] = (5 * alpha0 + 2 * alpha1) / 7
                        alphas[4] = (4 * alpha0 + 3 * alpha1) / 7
                        alphas[5] = (3 * alpha0 + 4 * alpha1) / 7
                        alphas[6] = (2 * alpha0 + 5 * alpha1) / 7
                        alphas[7] = (1 * alpha0 + 6 * alpha1) / 7
                    else:
                        alphas[2] = (4 * alpha0 + 1 * alpha1) / 5
                        alphas[3] = (3 * alpha0 + 2 * alpha1) / 5
                        alphas[4] = (2 * alpha0 + 3 * alpha1) / 5
                        alphas[5] = (1 * alpha0 + 4 * alpha1) / 5
                        alphas[6] = 0
                        alphas[7] = 255
                    colors = [rgb565toColor(color0),
                              rgb565toColor(color1)]
                    colors += [((2 * colors[0][0] + colors[1][0]) / 3, (2 * colors[0][1] + colors[1][1]) / 3, (2 * colors[0][2] + colors[1][2]) / 3)]
                    colors += [((2 * colors[1][0] + colors[0][0]) / 3, (2 * colors[1][1] + colors[0][1]) / 3, (2 * colors[1][2] + colors[0][2]) / 3)]
                    for k in xrange(16):
                        pixel = colors[bits(pixels, k*2, (k*2)+2)]
                        pixel += (alphas[bits(alpha, k*3, (k*3)+3)],)
                        self.maps[i].putpixel((x+(k%4), y+(k/4)), pixel)
                    x += 4
                    if x >= self.width:
                        y += 4
                        x = 0

if __name__ == "__main__":
    import os
    if 0:
        # Batch conversion, for use with package.py
        for dirname in os.listdir(os.curdir):
            dirname = os.path.join(os.curdir, dirname)
            if not os.path.isdir(dirname): continue
            for fname in os.listdir(dirname):
                if fname[-6:] != 'raster': continue
                rim = RasterImage()
                fin = open(os.path.join(dirname, fname), 'rb')
                rim.read(fin)
                rim.maps[0].save(os.path.join(dirname, fname[:-6]+'png'))
    else:
        # Single file conversion
        import sys
        rim = RasterImage()
        if len(sys.argv) < 2:
            sys.stderr.write("Usage:\n")
            sys.stderr.write("python raster.py <raster file> [<image file>]\n")
            sys.exit(1)
        fin = open(sys.argv[1], 'rb')
        rim.read(fin)
        if len(sys.argv) < 3:
            fname = os.path.splitext(sys.argv[1])[0]
            for i, m in enumerate(rim.maps):
                m.save(fname+str(i)+".png")
        else:
            rim.maps[0].save(sys.argv[2])
